import numpy as np
import pytest

import dace
from dace.transformation.dataflow import (MapFusion, ReduceExpansion,
                                          TrivialMapElimination, Vectorization,
                                          WarpTiling)
from dace.transformation.interstate import (GPUTransformSDFG, HoistState,
                                            InlineSDFG, StateFusion)
from dace.transformation.subgraph import MultiExpansion, SubgraphFusion

from daceml.distributed.schedule import lower
from daceml.distributed.utils import find_map_containing, compile_and_call

@dace.program
def softmax_fwd(inp: dace.float32[2, 16, 128, 128]):
    rowmax = np.maximum.reduce(inp, axis=-1, keepdims=True)
    exp_arr = np.exp(inp - rowmax)
    rowsum = np.add.reduce(exp_arr, axis=-1, keepdims=True)
    return exp_arr / rowsum


# Numerically-stable version of softmax
def softmax(x):
    tmp_max = np.max(x, axis=-1, keepdims=True)
    tmp_out = np.exp(x - tmp_max)
    tmp_sum = np.sum(tmp_out, axis=-1, keepdims=True)
    return tmp_out / tmp_sum

# Get SDFG
sdfg = softmax_fwd.to_sdfg()

# Apply transformations
sdfg.apply_transformations_repeated(ReduceExpansion)
sdfg.simplify()
MultiExpansion.apply_to(sdfg, sdfg.node(0).nodes())
SubgraphFusion.apply_to(sdfg, sdfg.node(0).nodes())
sdfg.expand_library_nodes()
sdfg.simplify()
#sdfg.apply_transformations_repeated([TrivialMapElimination, MapFusion])
#sdfg.apply_transformations(GPUTransformSDFG)
#assert sdfg.apply_transformations(WarpTiling) == 1
#sdfg.apply_transformations_repeated([HoistState, InlineSDFG, StateFusion])
#sdfg.apply_transformations_repeated([TrivialMapElimination, MapFusion])
#if vector_length != 1:
#    sdfg.apply_transformations_repeated(
#        Vectorization, dict(vector_len=vector_length, preamble=False, postamble=False, strided_map=False))

to_tile = find_map_containing(sdfg, 'outer_fused')
lower(sdfg, {to_tile: [2, 2, 1]})

# Check correctness
inp = np.random.rand(2, 16, 128, 128).astype(np.float32)
reg_out = softmax(inp)

compile_and_call(sdfg,
                 inputs={'inp': inp.copy()},
                 expected_output=reg_out,
                 num_required_ranks=4)
print(f"Passed using schedule {scheme}!")
